/*-----------------------------------------------------------------------
		
	Inline Math library : Lookup table implementation
	
	* See imath.h for more information

	By Matt Barnett
		(see References in imath.h)
	
------------------------------------------------------------------------*/

#ifndef _IMATH_LUT_IMP_HEADER_223457642
#define _IMATH_LUT_IMP_HEADER_223457642

//------- float_lookuptable ------------------------------------
//
//	Before i begin.. Yes, i know that the c++ 
//  standard library has a lookup table template class.
//
//
//  This is a simple generic lookup table(lut) class
//  for functions that return a single float
//	and take a single float parameter
//	(GPG2 see ref 2) has a better lookuptable)
//	
//	this works well for sinf etc
//	
//.	With bounds checking its quite a bit slower.
//	Without it, i think a lookup maybe amount to
//	only a few cycles (WRONG!, in retrospect, see asm anal),
//  because looking up consists of
//	an inline (no fuction call) function
//	which performs a single divide, and 
//  uses the clever real2int to get an int
//	to index the array with
//
//  ------ design notes -----
//
//  i decided to forgo inheritance
//
//  (it would be nice to have a base class
//   lookuptable and inhert specific
//	 table types)
//  
//  templates would probably be a good idea
//
//  (but an int indexing table would be
//   quite different than a float "indexing"
//   one, and could be faster, so the template
//   would have to account for that)
//
//	In the end decided to make this a simple class
//  with inlined functions specific to functions of
//	the form
//
//  float somefunction(float someparam);
//
//	becomes...
// 
//  float somelut.lookup(someparam);
//
//  The other lut class, is specifically
//  for generating random numbers and is
//  pretty much hard-coded to that effect.
//
//  ---------- performance notes ---------------
//  --------------------------------------------
//
//	*test yourself using test_speed(int);
//	*test in both Debug and Release modes
//    (only take Release mode seriously, but
//		its usefull to see the difference between
//			debug & release)
//
//
//	Debug mode is still faster than parent functions
//	which are them selves slower than in Release mode.
//	
//  faster than parent tri functions
//  generally slower for sqrtf
//  faster than on-the fly random number generation
//		(not yet suited to rng, which really
//		 deserves its own lut class)
//
//  --------------------------------------------

//  ---------- mem usage notes ---------------
//  ------------------------------------------
//
//100 floats	400bytes			//maybe for some stuff
//1000 floats = 4k					//vagly accurate
//10000 floats = 40k etc			//pretty accurate
//	
//	Need more accuracy? 
//  (fix class before using more samples)
//  
//	1)
//
//	I've made that the upper limit for the number of 
//	samples. Not due to memory usage but processor
//  cache usage. (im guessing here, from accumlated knowledge)
//	We want to be sure that the whole table fits in the cache
//  so that its really fast. (this afik occurs automagically)
//	If the table is to big, the chances are that the processor
//  will thrash the table. Much like a gfx card will thrash textures
//  between system/texture memory if the application tries to use to
//  much texture memory.					

//  ------------------------------------------

//if you want a quick report when a lookup table constructs then define FLUT_VERBOSE
//if you want to remove the construction message, rem FLUT_VERBOSE out

#define FLUT_VERBOSE

class float_lookuptable
{
	public:

	//constructor provides everything
	float_lookuptable(int innumsamples,float instart,float inend,float (*function)(float) )
	{
		//store the pointer to the function for testing later

		parentfunction = function;
		num_samples = innumsamples;
		
		if (innumsamples > 10000)
		{
			printf("float_lookuptable::float_lookuptable() num_samples > 10000!! rtfm\n");
			exit(0);
		}


		start = instart;
		end = inend;
		samples = 0;
		samples = meNew float [num_samples];
		assert(samples);

		//may as well fabs() to be sure we get a positive
		//delta (as this is precalc
		float diff = fabs(end - start);
		delta = diff / (float)(num_samples);
	  
		float x = start;

		#ifdef FLUT_VERBOSE
			printf("float_lookuptable( start(%f),end(%f),delta(%f),num(%i) )\n",start,end,delta,num_samples);
		#endif

		for (int i = 0; i < num_samples; i++)
		{
			//assert(index < num_samples);
		
			x = start + (diff * (float)i / (num_samples) );
			samples[i] = (*parentfunction)(x);
		}
	}	

	//looks up like the function
	inline float lookup(float lookup_what) const
	{
		lookup_what /= delta; 
		int il = im_real2int(lookup_what);		//fast
		//int il = (int)lookup_what ;			//doesnt seem any slower
		
		//we only need to check the index is safe
		//to access the array.  These are faster
		//than comparing floats
		// (we could use a smaller precision index
		//  and possibly gain more speed)



		//was previously this line (which caused the assertion(il < num_samples) to fail (+3v!L />U|\|7i/\/\3 3/>/>0z)
		//while (il >= num_samples) il -= num_samples;
		while (il >= num_samples) il -= num_samples;
		while (il < 0) il += num_samples;
		
		assert(il >= 0);
		assert(il < num_samples);

		return samples[il];
	}

	
	void test_speed(int test_samples) const
	{	
		printf("\n\ntesting speed\n\n");

		//i think as long as we do enough samples
		//it will be an ok test

		long before,after,deltatime;
		
		float some_number = two_pi/1.23456789;		//get an angle
		float out;									//temp to store output of function
		int x = 0;									//counter

		before = GetTickCount();					//time before
	
		for (x=0;x<test_samples;x++)				//do the test
			out = lookup(some_number);
		
		after = GetTickCount();						//time after
		deltatime = after - before;					//delta

		printf("lookup(%f) x %i = (%f) in %i msec\n",some_number,samples,out,deltatime);

		before = GetTickCount();					//time before

		for (x=0;x<test_samples;x++)				//do the test
			out = parentfunction(some_number);

		after = GetTickCount();						//time after
		deltatime = after - before;					//delta


		printf("parentfunction(%f) x %i = (%f) in %i msec\n",some_number,samples,out,deltatime);
	}

	~float_lookuptable()
	{
		meDelete [] samples;
	}


	private:
			float *samples;
			int num_samples;
			float start,end;
			float delta;
			float (*parentfunction)(float);


	//no assignment or copy kids
	operator=(float_lookuptable &);
	float_lookuptable(float_lookuptable &);
};

//------------------------------------- 
//hard coded random number lookup table
//-------------------------------------

//note: this is slower in debug mode than
//      rand();
//
//      but in release mode it beats rand()
//      however, because rand()  is so fast
//      its probably not worth the memory

//quick prototype
void im_random_seed_time();

class float_random_lookuptable
{
	public:
		float_random_lookuptable(int numsamples)
		{
			//seed the generator
			im_random_seed_time();

			//class general
			num_samples = numsamples;
			current_index = 0;

			//alloc memory
			samples = 0;
			samples = meNew float [num_samples];
			assert(samples);

			//fill the array
			for (int x=0;x<num_samples;x++)
			{
				samples[x] = ((float)rand()/(RAND_MAX+1));
			}

		}
		~float_random_lookuptable()
		{
			meDelete [] samples;
		}

		inline float lookup_random_num()
		{
			return samples[current_index];
		
			current_index++;

			//loop around again
			if (current_index > num_samples) current_index = 0;
		
		}
			
		//not const becauase it calls lookup_random_num
		void test_speed(int test_samples)
		{	
			printf("\n\ntesting speed\n\n");

			//i think as long as we do enough samples
			//it will be an ok test

			long before,after,deltatime;

			float out;									//temp to store output of function
			int x = 0;									//counter

			before = GetTickCount();					//time before
		
			for (x=0;x<test_samples;x++)				//do the test
				out = lookup_random_num();
			
			after = GetTickCount();						//time after
			deltatime = after - before;					//delta

			printf("lookup_random_num() x %i = %i msec\n",test_samples,deltatime);

			before = GetTickCount();					//time before

			for (x=0;x<test_samples;x++)				//do the test
				out = ((float)rand()/(RAND_MAX+1));

			after = GetTickCount();						//time after
			deltatime = after - before;					//delta

			printf("((float)rand()/(RAND_MAX+1)) x %i = %i msec\n",test_samples,deltatime);
		}

	private:
		int num_samples,current_index;
		float *samples;
};

#endif