/*-----------------------------------------------------------------------
		
	Inline Math library : Matrix and Quaternion implementation
	
	* See imath.h for more information

	By Matt Barnett
		(see References in imath.h)
	
------------------------------------------------------------------------*/


#ifndef _IMATH_MQ_IMP_HEADER_4778097634
#define _IMATH_MQ_IMP_HEADER_4778097634

//----------------------------
//----------------------------

class iquaternion4
{
	public:
		
		//quaternion data (w,x,y,z)
		union
		{
			float q[4];	
			struct { float w, x, y, z; };
		};
		
		//-------------- operators ---
		//---------------------------- 
		inline float operator[](int subscript) 
		{
			assert( subscript < 4 && subscript >= 0 );
			return q[subscript];
		}

		//assignment
		inline const iquaternion4 &operator=(const iquaternion4 &right)
		{
			if (&right != this) {				//check for self assign
				w = right.w;
				x = right.x;					//copy right to left					
				y = right.y;	
				z = right.z;	
			}
			return *this;						//enables x = y = z						
		}
	
		//----------------------------
		//----------------------------
		
		//------- construction etc ---
		//----------------------------

		//default ctor
		inline iquaternion4()		
		{
			makeMultiplicationIdentity(); 
		}
		
		//ctor											
		inline iquaternion4(float setw,float setx,float sety,float setz)
		{
			set(setw,setx,sety,setz);
		}												
	
		//ctor											
		inline iquaternion4(float angle,ivector3 &axis)
		{
			makeAngleAxisRotation(angle,axis);
		}
		
		//ctor											
		explicit inline iquaternion4(ivector3 &euler)
		{
			makeEulerRotation(euler);
		}	

		//dtor(unnessesary)
		inline ~iquaternion4()	
		{
		}						
	
		//----------------------------
		//---------------------------- 

		//--------- void functions ---
		//----------------------------
		inline void set(float setw,float setx,float sety,float setz)
		{
			w = setw;
			x =	setx;
			y = sety;
			z = setz;
		}
		
		//dumps the quaternion componets
		inline const void cdump(const char *caption)  
		{
			cout << caption << w << " " << x << " " << y << " " << z << endl;
		}	

		//theroy (identity quaternion)
		//
		//There are two identity quaternions.
		//In this context, we only use the one that multiplies with another quaternion 
		//to give the same quaternion ( the multiplication identity )

		inline void makeMultiplicationIdentity()
		{
			q[0] = 1.0;		//w
			q[1] = 0.0;		//x
			q[2] = 0.0;		//y
			q[3] = 0.0;		//z
		}

		//from Ref 5. (* See imath.h)
		//(http://gamedev.net/reference/articles/article1691.asp#Q53)
		//Qr = Q1.Q2 = ( w1.w2 - v1.v2, w1.v2 + w2.v1 + v1 x v2 )
		//
		//where v1 = (x,y,z) of Q1
		//	w1 = (w)     of Q1
		//	v2 = (x,y,z) of Q2
		//	w2 = (w)     of Q2
		//
	    //and both . and x are the standard vector dot and cross products.
		//its easy(ish) to follow that the multiply function 
		//doesnt chane the quaternion if multiply is the  identity
		inline void multiply(const iquaternion4 &multiply)
		{
			iquaternion4 temp;
			temp.x = ((w * multiply.x) + (x * multiply.w) + (y * multiply.z) - (z * multiply.y));
			
			//for multiply == identity quaternion
			//x = 1 * x + 0 * w + 0 * y + 0 * z;
			//x = x

			temp.y = ((w * multiply.y) - (x * multiply.z) + (y * multiply.w) + (z * multiply.x));
			temp.z = ((w * multiply.z) + (x * multiply.y) - (y * multiply.x) + (z * multiply.w));
			temp.w = ((w * multiply.w) - (x * multiply.x) - (y * multiply.y) - (z * multiply.z));
			*this = temp;
		}

		
		inline void makeAngleAxisRotation(float angle,ivector3 &axis)
		{
			const float hangle(angle * 0.5f);
			const float sina(im_sin(hangle) );
			x = (axis.x * sina);
			y = (axis.y * sina);
			z = (axis.z * sina);
			w = cos(hangle);
		}

		//from Ref 5. (* See imath.h)
		//(more or less unchanged from what that articles has)
		//(http://gamedev.net/reference/articles/article1691.asp#Q60)
			//" The following more or less comes from:
			//  http://vered.rose.utoronto.ca/people/david_dir/GEMS/GEMS.html "

		inline void makeEulerRotation(ivector3 &angles)
		{
			//Pitch->X axis, Yaw->Y axis, Roll->Z axis
		
			const float sinPitch(im_sin(angles.y*0.5f));
			const float cosPitch(im_cos(angles.y*0.5f));

			const float sinYaw(im_sin(angles.z*0.5f));
			const float cosYaw(im_cos(angles.z*0.5f));

			const float sinRoll(im_sin(angles.x*0.5f));
			const float cosRoll(im_cos(angles.x*0.5f));
			const float cosPitch_cosYaw(cosPitch*cosYaw);
			const float sinPitch_sinYaw(sinPitch*sinYaw);

			x = sinRoll * cosPitch_cosYaw   - cosRoll * sinPitch_sinYaw;
			y = cosRoll * sinPitch * cosYaw + sinRoll * cosPitch * sinYaw;
			z = cosRoll * cosPitch * sinYaw - sinRoll * sinPitch * cosYaw;
			w = cosRoll * cosPitch_cosYaw   + sinRoll * sinPitch_sinYaw;
		}

		//this is the alternative
		//multiplying euler angles in the correct order
		//i _think_ that the above is simply a re-arrangment
		//but i think the ordering here is wrong or somthing
		//(whatever, the above works the way i want and this doesnt)

		//
		//inline void makeEulerRotation(ivector3 &angles)
		//{
		//	iquaternion4 qx( im_cos(angles.x/2),im_sin(angles.x/2),0,0 );
		//	iquaternion4 qy( im_cos(angles.y/2),0,im_sin(angles.y/2),0 );	
		//	iquaternion4 qz( im_cos(angles.z/2),0,0,im_sin(angles.z/2) );
		//	qx.multiply ( qy );
		//	qx.multiply ( qz );
		//	*this = qx;
		//}
		//
	
		//---------------------------- 
		//---------------------------- 

		//-------------- functions ---
		//----------------------------

		//well this seems to draw the expected vector
		//but its easier to extract the basis from the
		//matrix, since the quat is converted to one in the
		//end
		const ivector3 &direction()
		{
			ivector3 out;
			out.x = 2.0f * ( x * z - w * y );
			out.y = 2.0f * ( y * z + w * x );
			out.z = 1.0f - 2.0f * ( x * x - y * y );

			out.normalise();
			return out;
		}

		inline float dot(const iquaternion4 &right)
		{
			return ( (x * right.x) + (y * right.y) +
					 (z * right.z) + (w * right.w) );
		}
		inline float lengthsqr()
		{
			return dot(*this);
		}
		inline float length()
		{
			return sqrt( lengthsqr() );
		}

		//just like a vector ( UNTESTED )
		inline void normalise()
		{
			const float leng(length() );
			x /= leng;
			y /= leng;
			z /= leng;
			w /= leng;
		}

		//?? inline iquaternion4 &lerp(iquaternion4 &other,float t);
		//?? inline iquaternion4 &slerp(iquaternion4 &other,float t);	
		//?? inline void invert(iquaternion4 &multiply);

		//---------------------------- 
		//---------------------------- 
};

//-------------- imatrix16 -----------------------------------------------------------
//
//  ---------------------------------
//  Row and clomn major format issues
//  ---------------------------------
//
//	Currently, this is an OpenGL freindly matrix
//  class.  Direct3D stores matricies in a different
//  order to OpenGL, and this would need to be accounted
//  for if Direct3D was used.  Also, an engine might need
//  to use both formats if it supports both APIs.  This is
//  why i have differed implementing this feature untill i
//  know how to get it right.
//
//

//matrix in colomn major format (consistant with opengl)
// 
//		eg { m[0] m[4] m[8]  m[12] 
//		 	 m[1] m[5] m[9]  m[13]
//			 m[2] m[6] m[10] m[14]
//			 m[3] m[7] m[11] m[15] }
//

//note: it would be cool to have a union for the matrix data allowing sensible names
//      for different parts of the matrix.  For instance, the first 3x3 block represents
//      the rotation in the form of the orthonormal basis axes. The matrix also stores scaling etc.
//
//	    somthing like...
//
//							
//		union
//		{
//			float m[16];
//			struct { float col0[4],col1[4],col2[4],col3[4]; };
//			
//		};		
//
//---------------------------------------------------------------------------------

class imatrix16
{
	public:

		//------- construction etc ---
		//----------------------------
		float m[16];																						//matrix data
		inline imatrix16()										 { makeIdentity(); }						//ctor											
		inline imatrix16(iplane4 &cast_floor,ivector3 lightpos)  { makeProjection(cast_floor,lightpos); }	//ctor
		explicit inline imatrix16(iquaternion4 &createfrom)		 { makefromQuaternion(createfrom); }		//ctor

		inline ~imatrix16()	{}																			//dtor(unnessesary)
		//---------------------------- 
		//---------------------------- 

		//--------- void functions ---
		//----------------------------	
		
		//dumps the matrix componets
		inline const void cdump(const char *caption)  
		{
			cout << caption << " row major\n";

			cout << caption << " row 1 " << m[0] << " " << m[1] << " " << m[2] << " " << m[3] << endl;
			cout << caption << " row 2 " << m[4] << " " << m[5] << " " << m[6] << " " << m[7] << endl;
			cout << caption << " row 3 " << m[8] << " " << m[9] << " " << m[10] << " " << m[11] << endl;
			cout << caption << " row 4 " << m[2] << " " << m[13] << " " << m[14] << " " << m[15] << endl;
			
			cout << caption << " col major\n";

			cout << caption << " col 1 " << m[0] << " " << m[4] << " " << m[8] << " " << m[12] << endl;
			cout << caption << " col 2 " << m[1] << " " << m[5] << " " << m[9] << " " << m[13] << endl;
			cout << caption << " col 3 " << m[2] << " " << m[6] << " " << m[10] << " " << m[14] << endl;
			cout << caption << " col 4 " << m[3] << " " << m[7] << " " << m[11] << " " << m[15] << endl << endl;
		}	

		//inline void invert
		//inline void transpose
		inline void makeIdentity() 
		{		
			m[0] = 1.0;
			m[1] = 0.0;
			m[2] = 0.0;
			m[3] = 0.0;

			m[4] = 0.0;
			m[5] = 1.0; 
			m[6] = 0.0;
			m[7] = 0.0;

			m[8] = 0.0; 
			m[9] = 0.0; 
			m[10] = 1.0;
			m[11] = 0.0; 

			//last right hand colomn
			m[12] = 0.0; 
			m[13] = 0.0; 
			m[14] = 0.0; 
			m[15] = 1.0; 
		}

		//this makes a projection matrix
		//which can be fed into a graphics API
		//
		//is by no means perfect as a shadow/projection
		//matrix (still a work in progress)

		//from http://www.sgi.com/software/opengl/advanced96/node46.html
		inline void makeProjection(iplane4 &cast_floor,ivector3 &light_pos) 
		{
			float light_posw = 0.0;							//0.0?

			float dot = cast_floor.norm.x * light_pos.x +
					    cast_floor.norm.y * light_pos.y +
						cast_floor.norm.z * light_pos.z +
						cast_floor.d * light_posw;				

			m[0] = dot - light_pos.x * cast_floor.norm.x;
			m[1] = 0.0 - light_pos.y * cast_floor.norm.x;
			m[2] = 0.0 - light_pos.z * cast_floor.norm.x;
			m[3] = 0.0 - light_posw  * cast_floor.norm.x;
    
			m[4] = 0.0 - light_pos.x * cast_floor.norm.y;
			m[5] = dot - light_pos.y * cast_floor.norm.y;
			m[6] = 0.0 - light_pos.z * cast_floor.norm.y;
			m[7] = 0.0 - light_posw  * cast_floor.norm.y;
    
			m[8 ] = 0.0 - light_pos.x * cast_floor.norm.z;
			m[9 ] = 0.0 - light_pos.y * cast_floor.norm.z;
			m[10] = dot - light_pos.z * cast_floor.norm.z;
			m[11] = 0.0 - light_posw  * cast_floor.norm.z;
    
			m[12] = 0.0 - light_pos.x * cast_floor.d;
			m[13] = 0.0 - light_pos.y * cast_floor.d;
			m[14] = 0.0 - light_pos.z * cast_floor.d;
			m[15] = dot - light_posw  * cast_floor.d;
		} 
	
		
		//from gametutorials.net (see ref 6)
		inline void makefromQuaternion(iquaternion4 &q)		 
		{
			
			// First row
			m[ 0] = 1.0f - 2.0f * ( q.y * q.y + q.z * q.z );  
			m[ 1] = 2.0f * ( q.x * q.y - q.w * q.z );  
			m[ 2] = 2.0f * ( q.x * q.z + q.w * q.y );  
			m[ 3] = 0.0f;  

			// Second row
			m[ 4] = 2.0f * ( q.x * q.y + q.w * q.z );  
			m[ 5] = 1.0f - 2.0f * ( q.x * q.x + q.z * q.z );  
			m[ 6] = 2.0f * ( q.y * q.z - q.w * q.x );  
			m[ 7] = 0.0f;  

			// Third row
			m[ 8] = 2.0f * ( q.x * q.z - q.w * q.y );  
			m[ 9] = 2.0f * ( q.y * q.z + q.w * q.x );  
			m[10] = 1.0f - 2.0f * ( q.x * q.x + q.y * q.y );  
			m[11] = 0.0f;  

			// Fourth row
			m[12] = 0;  
			m[13] = 0;  
			m[14] = 0;  
			m[15] = 1.0f;
		}

		//this functions could be avoided by
		//providing better access to the matrix's
		//internal values
		//based on theroy (Q5 of matrix FAQ) (see ref 5)
		inline ibasis9 &getBasis()
		{
			//prepare the basis based on the matrix
			ivector3 xa( m[0],m[1],m[2] );
			ivector3 ya( m[4],m[5],m[6] );
			ivector3 za( m[8],m[9],m[10] );
			//make it and return it
			ibasis9 out(xa,ya,za);
	
			return out;
		}

		//
		void multiply_vectors(ivector3 *vector_array,int num_vectors)
		{
			assert(vector_array);
			assert(num_vectors == 1);

			ivector3 out;
			out.x = m[0] * vector_array->x + m[4] * vector_array->y + m[8] * vector_array->z;
			out.y = m[1] * vector_array->x + m[5] * vector_array->y + m[9] * vector_array->z;
			out.z = m[2] * vector_array->x + m[6] * vector_array->y + m[10] * vector_array->z;
			*vector_array = out;
		}	

		//??? inline void makePerspectiveProjection() {} 
		//??? inline void makeOrthographicProjection() {}
		//??? inline void makeEulerRotation(float angle, ivector3 &axis) {}
		//??? inline void makeAngleAxisRotation(float angle, ivector3 &axis) {}

		//---------------------------- 
		//---------------------------- 
};

#endif
