/*-----------------------------------------------------------------------
		
	Inline Math library : Ray and Plane implementation
	
	* See imath.h for more information

	By Matt Barnett
		(see References in imath.h)
	
------------------------------------------------------------------------*/

#ifndef _IMATH_TRACE_IMP_HEADER_529081235
#define _IMATH_TRACE_IMP_HEADER_529081235

/*---------------------------------------------
	
	Raytracing/linetracing and collision detection

	iray6
	iplane6

	-------
	classes
	-------
		
		iray6
		iplane6

	------
	design
	------

		* a ray is defined a point(position) and a normalised vector(direction)
		* a plane is defined as abcd
	
---------------------------------------------*/

//-------------- iplane4 ------------
//	plane class
//------------------------------------

class iplane4
{
	public:
												
		ivector3 norm;
		float d;
		
		//-------------- operators ---
		//---------------------------- 
	    //return types according to Effective C++ 2e (see ref 8)

		//assignment
		inline const iplane4 &operator=(const iplane4 &right)
		{
			if (&right != this) {				//check for self assign
				norm = right.norm;				//copy right to left					
				d = right.d;		
			}
			return *this;						//enables x = y = z						
		}

		//---------------------------- 
		//---------------------------- 

		//------- construction etc ---
		//----------------------------
		
		//inlined ctor			
		inline iplane4()						   
		{
			//no need here
			//vectors construct themselves adequetly
		}					
		
		//inlined ctor	
		inline iplane4(const float in_d,const ivector3 &in_norm)
		{
			norm = in_norm;
			d = in_d;
		}					
		
		//inlined dtor(unnessesary)
		inline ~iplane4()
		{}	
		
		//---------------------------- 
		//---------------------------- 

		//--------- functions --------
		//----------------------------
		//sets components
													
		//dumps the componets (debug aid)
		inline void cdump(const char *caption)  const 
		{
			norm.cdump("iplane4 norm = ");
			printf("iplane4 d = %f\n",d);
		}	
		
		//draw's the plane (debug aid)
		//plane extend to infinity,
		//so the client decideds how far that is (length)
		//(not optimised for drawing lots of planes, which may be needed later)

		//creates a quad from the plane's normal
		//much like a particle is generated7
		inline void gldraw(float length,ivector3 rgbcolor)			
		{
			//some good practice at vector math here
			
			//we need to create an up vector and a left vector
			//from the normal of the plane

			//both are perpendicular to the norm, so crossproducts
			//are an easy way to do this (allthought there may
			//be a faster solution)

			//create any vector perp to the normal
			//(the operand xaxis could be any vector)
			ivector3 left = norm.cross(xaxis);
			left.normalise();
			
			//the up vector needs to be perp to both the norm and the left vector
			ivector3 up =  norm.cross(left);
			up.normalise();
			
			//size it according to length param
			left.multiply(length);
			up.multiply(length);

			//pos is a point on the plane
			ivector3 pos = (norm * d);
			ivector3 temp = pos;

			glDisable(GL_LIGHTING);

			glBegin(GL_LINES);

				glColor3fv(red.v);

				pos.glplot();temp = pos + norm;
				temp.glplot();		
	
				glColor3fv(cyan.v);
			
				pos.glplot();temp = pos + left;
				temp.glplot();

				glColor3fv(yellow.v);

				pos.glplot();temp = pos + up;
				temp.glplot();

			glEnd();

			glColor3fv(rgbcolor.v);

			glBegin(GL_QUADS);
				glVertex3f(pos.x + left.x + up.x,pos.y + left.y + up.y,pos.z + left.z + up.z);
				glVertex3f(pos.x - left.x + up.x,pos.y - left.y + up.y,pos.z - left.z + up.z);
				glVertex3f(pos.x - left.x - up.x,pos.y - left.y - up.y,pos.z - left.z - up.z);
				glVertex3f(pos.x + left.x - up.x,pos.y + left.y - up.y,pos.z + left.z - up.z);
			glEnd();
		
		}		
 
		//---------------------------- 
		//----------------------------

};

//-------------- iray ------------
//
//	ray class
//------------------------------------

class iray6
{
	public:
												
		ivector3 pos;		//treated as a position
		ivector3 dir;		//treated as a direction (magnitude represents light/strength?)
		

		//-------------- operators ---
		//---------------------------- 
	    //return types according to Effective C++ 2e (see ref 8)

		//assignment
		//the comilier should be able to generate this (its only member copy)
		inline const iray6 &operator=(const iray6 &right)
		{
			if (&right != this) {				//check for self assign
				dir = right.dir;				//copy right to left					
				pos = right.pos;		
			}
			return *this;						//enables x = y = z						
		}

		//---------------------------- 
		//---------------------------- 

		//------- construction etc ---
		//----------------------------
		
		//inlined ctor			
		inline iray6()						   
		{
			pos.set(0,0,0);			//we'll start all rays at 0,0,0
			dir.set(0,0,1.0);		//pointing up in the air
		}					
		
		//inlined ctor(initalises to pos,dir)	
		inline iray6(const ivector3 &in_pos,const ivector3 &in_dir)
		{
			pos = in_pos;
			dir = in_dir;
		}	
		
		//inlined cpy ctor
		//the comilier generates this (its only member copy)
		
		//inlined dtor(unnessesary)
		inline ~iray6()
		{}	
		
		//---------------------------- 
		//---------------------------- 

		//--------- functions --------
		//----------------------------

		inline float intersect(class iplane4 &with)
		{
			
			float cosAlpha = dir.dot( with.norm );		//dot product of ray's direction vector (normalised) and plane's normal vector (normalised)
		
			ivector3 diff = (with.norm * with.d) - pos;				//vector from plane origin to ray origin
			float deltaD = diff.dot(with.norm);				//dot product of above and plane's normal
			
			//deltaD is the length of the shortest line to the point 

			return deltaD/cosAlpha;
		}


		//dumps the componets (debug aid)
		inline void cdump(const char *caption)  const 
		{
			pos.cdump("ray pos = ");
			dir.cdump("ray dir = ");
		}	
		
		//draw's the ray (debug aid)
		//rays extend to infinity,
		//so the client decideds how far that is (length)

		//(not optimised for drawing lots of rays, which may be needed later)		
		inline void gldraw(float length,const ivector3 &startcol,const ivector3 &endcol) const			
		{
			ivector3 endpos = dir;
				
			endpos.multiply(length);
			endpos.add(pos);

			glDisable(GL_LIGHTING);		
			//glDisable(GL_BLEND);
			glDisable(GL_TEXTURE_2D);
			glBegin(GL_LINES);
				glColor3fv(startcol.v);
				pos.glplot();
				glColor3fv(endcol.v);
				endpos.glplot();
			glEnd();
		}	
		
		inline void gldraw(float length) const			
		{
			ivector3 endpos = dir;
			endpos.multiply(length);
			endpos.add(pos);
			glDisable(GL_LIGHTING);
			glBegin(GL_LINES);
				pos.glplot();
				endpos.glplot();
			glEnd();
		}	
 
		//---------------------------- 
		//----------------------------

};



#endif